# SPDX-FileCopyrightText: 2021 Lukas Schrangl <lukas.schrangl@tuwien.ac.at>
#
# SPDX-License-Identifier: BSD-3-Clause

import enum
import traceback
from typing import Any, Callable, List, Optional, Union

from PyQt5 import QtCore, QtQuick, QtQml

from .dataset import DatasetCollection
from .item_models import ListModel
from .thread_worker import ThreadWorker


class BatchWorker(QtQuick.QQuickItem):
    """QtQuick item displaying progress bar while doing work in another thread

    This is useful when some calculation should be done with each entry of
    a dataset.
    """
    class ErrorPolicy(enum.IntEnum):
        """What to do if an error occurs while processing a dataset entry."""
        Abort = 0
        Continue = enum.auto()

    QtCore.Q_ENUM(ErrorPolicy)

    def __init__(self, parent: Optional[QtQuick.QQuickItem] = None):
        """Parameters
        ----------
        parent
            Parent QQuickItem
        """
        super().__init__(parent)
        self._dataset = ListModel()
        self._func = None
        self._worker = None
        self._argRoles = []
        self._kwargRoles = []
        self._resultRole = ""
        self._displayRole = ""
        self._count = -1
        self._progress = 0
        self._curIndex = -1
        self._curDsetIndex = -1
        self._curDset = None
        self._errorPolicy = self.ErrorPolicy.Abort
        self._errLst = []

    datasetChanged = QtCore.pyqtSignal()
    """:py:attr:`dataset` was changed"""

    @QtCore.pyqtProperty(QtCore.QVariant, notify=datasetChanged)
    def dataset(self) -> ListModel:
        """Apply :py:attr:`func` to each entry of this dataset. If this is
        a :py:class:`DatasetCollection` instance, :py:attr:`func` is applied
        to each entry of each element.
        """
        return self._dataset

    @dataset.setter
    def dataset(self, data: ListModel):
        if data is self._dataset:
            return
        self._dataset = data
        self.datasetChanged.emit()

    funcChanged = QtCore.pyqtSignal()
    """:py:attr:`func` was changed"""

    @QtCore.pyqtProperty(QtCore.QVariant, notify=funcChanged)
    def func(self) -> Callable:
        """Function to apply to each dataset entry. This should take each
        of :py:attr:`argRoles` as a keyword argument. The return value is
        stored in :py:attr:`dataset` using :py:attr:`resultRole`.
        """
        return self._func

    @func.setter
    def func(self, f: Callable):
        if self._func is f:
            return
        self._func = f
        self.funcChanged.emit()

    argRolesChanged = QtCore.pyqtSignal()
    """:py:attr:`argRoles` was changed"""

    @QtCore.pyqtProperty(list, notify=argRolesChanged)
    def argRoles(self) -> List[str]:
        """For each dataset entry, these roles are passed as positional
        arguments to :py:attr:`func`.
        """
        return self._argRoles

    @argRoles.setter
    def argRoles(self, r: List[str]):
        if r == self._argRoles:
            return
        self._argRoles = r
        self.argRolesChanged.emit()

    kwargRolesChanged = QtCore.pyqtSignal()
    """:py:attr:`kwargRoles` was changed"""

    @QtCore.pyqtProperty(list, notify=kwargRolesChanged)
    def kwargRoles(self) -> List[str]:
        """For each dataset entry, these roles are passed as keyword arguments
        to :py:attr:`func`.
        """
        return self._kwargRoles

    @kwargRoles.setter
    def kwargRoles(self, r: List[str]):
        if r == self._kwargRoles:
            return
        self._kwargRoles = r
        self.kwargRolesChanged.emit()

    resultRoleChanged = QtCore.pyqtSignal()
    """:py:attr:`resultRole` was changed"""

    @QtCore.pyqtProperty(str, notify=resultRoleChanged)
    def resultRole(self) -> str:
        """Store the results of :py:attr:`func` calls in :py:attr:`datasets`
        using this role. If it does not exist, it is created. If empty,
        results are not stored.
        """
        return self._resultRole

    @resultRole.setter
    def resultRole(self, r: str):
        if r == self._resultRole:
            return
        self._resultRole = r
        self.resultRoleChanged.emit()

    displayRoleChanged = QtCore.pyqtSignal()
    """:py:attr:`displayRole` was changed"""

    @QtCore.pyqtProperty(str, notify=displayRoleChanged)
    def displayRole(self) -> str:
        """Role to use for displaying currently processed item"""
        return self._displayRole

    @displayRole.setter
    def displayRole(self, r: str):
        if self._displayRole == r:
            return
        self._displayRole = r
        self.displayRoleChanged.emit()

    countChanged = QtCore.pyqtSignal()
    """:py:attr:`count` was changed"""

    @QtCore.pyqtProperty(int, notify=countChanged)
    def count(self) -> int:
        """Number of dataset entries to be processed. This is only updated
        on calling :py:meth:`start`.
        """
        return self._count

    progressChanged = QtCore.pyqtSignal()
    """:py:attr:`progress` was changed"""

    @QtCore.pyqtProperty(int, notify=progressChanged)
    def progress(self) -> int:
        """Number of processed dataset entries"""
        return self._progress

    errorPolicyChanged = QtCore.pyqtSignal()
    """:py:attr:`errorPolicy` was changed"""

    @QtCore.pyqtProperty(ErrorPolicy, notify=errorPolicyChanged)
    def errorPolicy(self) -> ErrorPolicy:
        return self._errorPolicy

    @errorPolicy.setter
    def errorPolicy(self, ep: ErrorPolicy):
        if self._errorPolicy == ep:
            return
        self._errorPolicy = ep
        self.errorPolicyChanged.emit()

    _errorListChanged = QtCore.pyqtSignal()
    """:py:attr:`_errorList` was changed"""

    @QtCore.pyqtProperty(list, notify=_errorListChanged)
    def _errorList(self) -> Union[List[str], List[int]]:
        """Data items for which errors were encountered.

        If :py:attr:`displayRole` was set, this contains the corresponding
        entries, otherwise indices are used.
        """
        return self._errLst

    isRunningChanged = QtCore.pyqtSignal()
    """:py:attr:`isRunning` was changed"""

    @QtCore.pyqtProperty(bool, notify=isRunningChanged)
    def isRunning(self):
        return self._worker is not None and self._worker.enabled

    _currentItemChanged = QtCore.pyqtSignal()
    """:py:attr:`_currentItem` was changed"""

    @QtCore.pyqtProperty(str, notify=_currentItemChanged)
    def _currentItem(self) -> str:
        """Currently processed item to be displayed beneath progress bar"""
        if (not self._displayRole or self._curDset is None or
                self._curIndex < 0):
            return ""
        return self._curDset.get(self._curIndex, self._displayRole)

    @QtCore.pyqtSlot()
    def start(self):
        """Start processing the data

        This updates :py:attr:`count`. After processing each dataset entry,
        :py:attr:`progress` is increased by 1. If ``progress == count``,
        processing has finished.
        """
        if isinstance(self._dataset, DatasetCollection):
            cnt = sum(self._dataset.get(i, "dataset").count
                      for i in range(self._dataset.count))
            # TODO: Handle empty
            self._curDset = self._dataset.get(0, "dataset")
        else:
            cnt = self._dataset.count
            self._curDset = self._dataset
        self._curIndex = 0
        self._curDsetIndex = 0

        self._errLst = []
        self._errorListChanged.emit()

        if cnt != self._count:
            self._count = cnt
            self.countChanged.emit()

        self._worker = ThreadWorker(self._func, enabled=True)
        self._worker.finished.connect(self._workerFinished)
        self._worker.error.connect(self._workerError)
        self.isRunningChanged.emit()

        if self._progress > 0:
            self._progress = 0
            self.progressChanged.emit()

        self._nextCall()

    @QtCore.pyqtSlot()
    def abort(self):
        """Abort processing"""
        self._curIndex = -1
        self._curDsetIndex = -1
        self._curDset = None
        if self._worker is None:
            return
        self._worker.enabled = False
        self._worker = None
        self.isRunningChanged.emit()

    def _nextCall(self):
        """Process next dataset entry"""
        while self._curIndex >= self._curDset.rowCount():
            self._curDsetIndex += 1
            self._curIndex = 0
            self._curDset = self._dataset.get(self._curDsetIndex, "dataset")
        self._currentItemChanged.emit()

        args = [self._curDset.get(self._curIndex, r) for r in self._argRoles]
        kwargs = {r: self._curDset.get(self._curIndex, r)
                  for r in self._kwargRoles}
        self._worker(*args, **kwargs)

    def _workerFinished(self, retval: Any):
        """Worker has finished

        Store the result in the dataset

        Parameters
        ----------
        retval
            Return value of :py:attr:`func` call
        """
        if self._resultRole:
            if self._resultRole not in self._curDset.roles:
                self._curDset.roles = self._curDset.roles + [self._resultRole]
            self._curDset.set(self._curIndex, self._resultRole, retval)
        self._progress += 1
        self._curIndex += 1
        self.progressChanged.emit()
        if self.progress < self._count:
            self._nextCall()
        else:
            self.abort()

    def _workerError(self, exc):
        tb = traceback.TracebackException.from_exception(exc)
        print("".join(tb.format()))

        self._errLst.append(self._currentItem or self._progress + 1)
        self._errorListChanged.emit()

        if self._errorPolicy == self.ErrorPolicy.Continue:
            self._progress += 1
            self._curIndex += 1
            self.progressChanged.emit()
            if self.progress < self._count:
                self._nextCall()
            else:
                self.abort()
        else:
            self.abort()


QtQml.qmlRegisterType(BatchWorker, "SdtGui.Templates", 0, 1, "BatchWorker")
